/* ffi.c - generate .ffi file from .h file
   Copyright (C) 2001, Clozure Associates.

This file is part of GNU CC (or, more accurately, part of a work
derived from GNU CC.)

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <ctype.h>
#include "config.h"
#include "input.h"
#include "tree.h"
#include "c-tree.h"
#include "flags.h"


FILE *ffifile;

int ffi_initialized = 0;

struct ffi_input_file {
  char *pathname;
  struct ffi_input_file *prev;
};

struct ffi_input_file *current_ffi_input_file = NULL;
char *ffi_current_source_file;

static int ffi_indent_count = 1;

void
ffi_debug_start_source_file (path)
     char *path;
{
  struct ffi_input_file *savef = 
    (struct ffi_input_file *) xmalloc(sizeof (struct ffi_input_file));

  ffi_indent_count = 1;
  savef->pathname = ffi_current_source_file;
  savef->prev = current_ffi_input_file;
  current_ffi_input_file = savef;
  ffi_current_source_file = (char *) xstrdup (path);
}

ffi_debug_end_source_file (lineno)
     unsigned lineno;
{
  struct ffi_input_file *currf = current_ffi_input_file;
  current_ffi_input_file = currf->prev;
  free (ffi_current_source_file);
  ffi_current_source_file = currf->pathname;
  free (currf);
}

void ffi_indent()
{
  int i;
  
  fprintf (ffifile, "\n");

  for (i = 0; i < ffi_indent_count; i++) {
    fprintf (ffifile, " ");
  }
  ffi_indent_count++;
}

void
ffi_unindent(n)
     int n;
{
  ffi_indent_count -= n;
}



enum ffi_typestatus {FFI_TYPE_UNDEFINED = 0,
		     FFI_TYPE_PRIMITIVE,
		     FFI_TYPE_TYPEDEF,
		     FFI_TYPE_STRUCT,
		     FFI_TYPE_UNION,
		     FFI_TYPE_ENUM,
		     FFI_TYPE_POINTER,
		     FFI_TYPE_FUNCTION,
		     FFI_TYPE_ARRAY,
		     FFI_TYPE_FIELD};

struct ffi_typeinfo {
  enum ffi_typestatus status;
  int type_number;
  tree type;
  char *name;
  char *source_file;
  unsigned source_line;
  int ref_type_number;		/* For pointer & array types, at least */
  int array_size;
  int first_field;
};

struct ffi_typeinfo *ffi_typevec;

int ffi_typevec_len;
int next_ffi_type_number = 1;

typedef struct {
  char *gcc_name;
  char *ffi_name;
} ffi_primitive_name_map;

ffi_primitive_name_map ffi_primitive_names[] =
{
  {"int", "int"},
  {"char", "char"},
  {"float", "float"},
  {"double", "double"},
  {"long double", "long-double"},
  {"void", "void"},
  {"long int", "long"},
  {"unsigned int", "unsigned"},
  {"long unsigned int", "unsigned-long"},
  {"long long int", "long-long"},
  {"long long unsigned int", "unsigned-long-long"},
  {"short int", "short"},
  {"short unsigned int", "unsigned-short"},
  {"signed char", "signed-char"},
  {"unsigned char", "unsigned-char"},
  {"complex int", "complex-int"},
  {"complex float", "complex-float"},
  {"complex double", "complex-double"},
  {"complex long double", "complex-long-double"},
  {NULL, NULL}
};

char *
ffi_primitive_type_name (decl)
     tree decl;
{
  ffi_primitive_name_map *map = ffi_primitive_names;
  char * gcc_name = IDENTIFIER_POINTER (DECL_NAME (decl));

  while (map->gcc_name != NULL) 
    {
      if (!strcmp (map->gcc_name, gcc_name))
	return map->ffi_name;
      map++;
    }

  fprintf (stderr, "Bug: couldn't find primitive name for %s\n",
	   gcc_name);
  exit (1);
}

struct ffi_typeinfo *
find_ffi_type_info (type)
     tree type;
{
  int index = TYPE_SYMTAB_ADDRESS (type);
  struct ffi_typeinfo * info;


  if (index == 0)
    {
      index = next_ffi_type_number++;
      TYPE_SYMTAB_ADDRESS (type) = index;
      
      if (next_ffi_type_number == ffi_typevec_len) 
	{
	  ffi_typevec 
	    = (struct ffi_typeinfo *) xrealloc (ffi_typevec,
						ffi_typevec_len * 2 *
						sizeof ffi_typevec[0]);
	  bzero ((char *) (ffi_typevec + ffi_typevec_len),
		 ffi_typevec_len * sizeof ffi_typevec[0]);
	  ffi_typevec_len *= 2;
	}
    }
  
  info = ffi_typevec + index;
  info->type_number = index;
  info->type = type;
  return info;
}

struct ffi_typeinfo *
ffi_create_type_info (type)
     tree type;
{
  struct ffi_typeinfo *info;
  tree lookup_type = type;

  if (TYPE_READONLY (type) || TYPE_VOLATILE (type))
    lookup_type = TYPE_MAIN_VARIANT (type);
  
  info = find_ffi_type_info (lookup_type);

  if (info->status != FFI_TYPE_UNDEFINED)
    return info;
 
  switch (TREE_CODE (type))
    {

    case POINTER_TYPE:
    case REFERENCE_TYPE:
      {
	tree pointer_target_type = TREE_TYPE (type);
	struct ffi_typeinfo *pointer_type_info;

	info->status = FFI_TYPE_POINTER;
	pointer_type_info = ffi_create_type_info (pointer_target_type);

	/* 'info' might have moved */
	info = find_ffi_type_info (type);
	info->ref_type_number = pointer_type_info->type_number;
      }
      break;

    case FUNCTION_TYPE:
      {
	tree arg_type;
	info->status = FFI_TYPE_FUNCTION;
	ffi_create_type_info (TREE_TYPE (type));
	for (arg_type = TYPE_ARG_TYPES (type); arg_type; arg_type = TREE_CHAIN (arg_type))
	  {
	    ffi_create_type_info (TREE_VALUE (arg_type));
	  }
	info = find_ffi_type_info (type);
      }
      break;

    case ARRAY_TYPE:
      {
	tree array_element_type = TREE_TYPE (type);
	struct ffi_typeinfo *element_type_info;
	int array_size_in_bits = 0, element_size_in_bits, array_size_in_elements;
	
	info->status = FFI_TYPE_ARRAY;
	element_type_info = ffi_create_type_info (array_element_type);
	info = find_ffi_type_info (type);
	info->ref_type_number = element_type_info->type_number;
	element_size_in_bits = TREE_INT_CST_LOW ( TYPE_SIZE (array_element_type));
	if (TYPE_SIZE (type))
	  array_size_in_bits = TREE_INT_CST_LOW ( TYPE_SIZE (type));
	array_size_in_elements = array_size_in_bits / element_size_in_bits;
	if (array_size_in_bits % element_size_in_bits)
	  {
	    fprintf (stderr, "screw : array size not integral\n");
	    exit (1);
	  }
	info->array_size = array_size_in_elements;
      }
      break;
      
    case RECORD_TYPE:
    case UNION_TYPE:
    case ENUMERAL_TYPE:
      {
	tree field;
	tree name = TYPE_NAME (type);
	if (name)
	  {
	    if (TREE_CODE (name) == IDENTIFIER_NODE)
	      info->name = (char *) xstrdup (IDENTIFIER_POINTER (name));
	    else if (TREE_CODE (name) == TYPE_DECL
		     && DECL_NAME (name))
	      info->name = (char *) xstrdup (IDENTIFIER_POINTER (DECL_NAME (name)));
	  }
	if (! info->name)
	  {
	    char buf[16];
	    sprintf(buf, "%d", info->type_number);
	    info->name = (char *) xstrdup (buf);
	  }
	if (TREE_CODE (type) == ENUMERAL_TYPE)
	  {
	    info->status = FFI_TYPE_ENUM;
	    break;
	  }

	info->status = (TREE_CODE (type) == RECORD_TYPE) ? FFI_TYPE_STRUCT : FFI_TYPE_UNION;

	for (field = TYPE_FIELDS (type); field; field = TREE_CHAIN (field))
	  {
	    ffi_create_type_info (TREE_TYPE (field));
	  }
	info = find_ffi_type_info (type);
      }
      break;


    default:
      fprintf (stderr, "Need to create info for type:\n");
      debug_tree (type);
      exit (1);
    }
  return info;
}

void
emit_ffi_type_reference (target_type)
     tree target_type;
{
  struct ffi_typeinfo *info;
  tree lookup_type = target_type;

  if (TYPE_READONLY (target_type) || TYPE_VOLATILE (target_type))
    lookup_type = TYPE_MAIN_VARIANT (target_type);

  info = ffi_create_type_info (lookup_type);

  switch (info->status) 
    {
    case FFI_TYPE_PRIMITIVE:
      fprintf (ffifile, "(%s ())", info->name);	/* No quotes around primitive name. */
      return;

    case FFI_TYPE_TYPEDEF:
      fprintf (ffifile, "(typedef \"%s\")", info->name);
      return;

    case FFI_TYPE_FUNCTION:
      {
	tree arg, arg_type;
	struct ffi_typeinfo *arg_type_info;
	/*
	  It seems like functions that take a fixed number of arguments
	  have a "void" argument after the fixed arguments, while those
	  that take an indefinite number don't.
	  That's perfectly sane, but the opposite of what LCC does.
	  So, if there's a "void" argument at the end of the arglist,
	  don't emit it; if there wasn't, emit one.
	  Sheesh.
	*/
	int had_void_arg = 0, args_emitted = 0;
	fprintf (ffifile, "(function");
	ffi_indent ();
	fprintf (ffifile, "(");
	for (arg = TYPE_ARG_TYPES (target_type); arg ; arg = TREE_CHAIN (arg))
	  {
	    if (args_emitted)
	      fprintf (ffifile, " ");
	    arg_type = TREE_VALUE (arg);
	    if (TREE_CODE (arg_type) != VOID_TYPE) {
	      emit_ffi_type_reference (arg_type);
	      args_emitted = 1;
	    }
	    else 
	      {
		had_void_arg = 1;
		break;
	      }
	  }
	if (! had_void_arg) {
	  if (args_emitted)
	    fprintf (ffifile, " ");
	  fprintf (ffifile, "(void ())");
	}
	fprintf (ffifile, ")");
	ffi_unindent (1);
	ffi_indent ();
	arg_type = TREE_TYPE (target_type);
	emit_ffi_type_reference (arg_type);
	ffi_unindent (1);
	fprintf (ffifile, ")");
      }
      break;
	
    case FFI_TYPE_POINTER:
      {
	tree pointer_target = TREE_TYPE (target_type);

	fprintf (ffifile, "(pointer ");
	emit_ffi_type_reference (pointer_target);
	fprintf (ffifile, ")");
	return;
      }
      
    case FFI_TYPE_STRUCT:
      fprintf (ffifile, "(struct-ref \"%s\")", info->name);
      break;

    case FFI_TYPE_UNION:
      fprintf (ffifile, "(union-ref \"%s\")", info->name);
      break;

    case FFI_TYPE_ENUM:
      fprintf (ffifile, "(enum-ref \"%s\")", info->name);
      break;

    case FFI_TYPE_ARRAY:
      {
	tree element_type = TREE_TYPE (target_type);

	fprintf (ffifile, "(array %d ", info->array_size);
	emit_ffi_type_reference (element_type);
	fprintf (ffifile, ")");
      }
      break;

    default:
      fprintf (stderr, "need to emit reference to this type\n");
      debug_tree (target_type);
      exit (1);
    }
}

void
ffi_emit_type_decl (info, decl)
     struct ffi_typeinfo *info;
     tree decl;
{
  struct ffi_typeinfo *target_info;
  tree target;

  if (TREE_ASM_WRITTEN (decl)) 
    return;

  TREE_ASM_WRITTEN (decl) = 1;
  fprintf (ffifile, "(type (\"%s\" %d)", info->source_file, info->source_line);
  ffi_indent ();
  fprintf (ffifile, "\"%s\"", info->name);
  ffi_unindent (1);
  target = DECL_ORIGINAL_TYPE (decl);
  ffi_indent ();
  emit_ffi_type_reference (target);
  ffi_unindent (1);
  fprintf (ffifile, ")\n");
}

char *
ffi_decl_name (decl)
     tree decl;
{
  if (DECL_NAME (decl))
    return IDENTIFIER_POINTER (DECL_NAME (decl));
  else
    return "";
}
void
ffi_emit_integer_constant (val)
     tree val;
{
  if (TREE_CODE (val) != INTEGER_CST) 
    {
      fprintf (stderr, "Expected integer_cst, got:\n");
      debug_tree (val);
      exit (1);
    }
  if (TREE_INT_CST_HIGH (val) == 0)
    fprintf (ffifile, HOST_WIDE_INT_PRINT_UNSIGNED, TREE_INT_CST_LOW (val));
  else if (TREE_INT_CST_HIGH (val) == -1
	   && TREE_INT_CST_LOW (val) != 0)
    {
      fprintf (ffifile, "-");
      fprintf (ffifile, HOST_WIDE_INT_PRINT_UNSIGNED,
	       -TREE_INT_CST_LOW (val));
    }
  else
    fprintf (ffifile, HOST_WIDE_INT_PRINT_DOUBLE_HEX,
	     TREE_INT_CST_HIGH (val), TREE_INT_CST_LOW (val));
}

void
ffi_emit_enum_values (val)
     tree val;
{
  fprintf (ffifile, "(");
  while (val)
    {
      fprintf (ffifile, "(\"%s\" ", IDENTIFIER_POINTER (TREE_PURPOSE (val)));
      ffi_emit_integer_constant (TREE_VALUE (val));
      fprintf (ffifile, ")");
      val = TREE_CHAIN (val);
    }
  fprintf (ffifile, ")");
}

void
ffi_herald (decl, prefix)
     tree decl;
     char *prefix;
{

  fprintf (ffifile, "(%s (\"%s\" %d)", prefix,
	   DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl));
}

void
ffi_emit_enum_idents (val)
     tree val;
{
  while (val)
    {
      fprintf (ffifile, "(enum-ident (\"\" 0)");
      ffi_indent ();
      fprintf (ffifile, "\"%s\" ", IDENTIFIER_POINTER (TREE_PURPOSE (val)));
      ffi_emit_integer_constant (TREE_VALUE (val));
      ffi_unindent (1);
      fprintf(ffifile, ")\n");
      val = TREE_CHAIN (val);
    }
}

/*
  It's hard to know exactly how to translate bit-fields.
  LCC only allows bit-fields to be of type "int" or "unsigned" (and
  doesn't even allow aliases to these types to be used.)
  LCC-FFIGEN therefore generates something like:

  ("x" (bit-field ("<type>" ()) <start-byte> <size-in-bytes> <a> <b> <c> <d>))

  where :
  <type> is one of int, unsigned
  <start-byte, size-in-bytes> refer to a properly-aligned fullword that
    "contains" the bit-field
  <a> is what lcc calls the "fieldleft" of the bit-field
  <b> is what lcc calls the "fieldright" of the bit-field
  <c> is the width of the bit-field
  <d> is a mask: (1- (ash 1 <c>))

  Hmm.  That doesn't have too much to do with the information that GCC
  provides, but we could probably fake it.  It seems silly to discard
  the information that GCC -does- provide, so let's instead generate:

  ("x" (bitfield ("<type>" ()) <start-bit> <size-in-bits>))

  where <start-bit> is the field position starting in bits, of 
  the bit closest to the beginning of the structure.

  Note that a GCC FIELD_DECL node X denotes a bit-field iff
   DECL_BIT_FIELD (X) is true and DECL_SIZE (X) is non-null.
*/

void
ffi_emit_field_list (field)
     tree field;
{
  int bit_offset, size_in_bits, is_bit_field, count = 0;
  tree field_type, size_node;
  char *field_name;

  ffi_indent ();
  fprintf (ffifile, "(");
  while (field)
    {
#if 0
      fprintf (stderr, "\nEmit field:\n");
      debug_tree (field);
#endif
      if (count > 0) 
	ffi_indent ();
      field_name = ffi_decl_name (field);
      is_bit_field = DECL_BIT_FIELD (field) && DECL_SIZE (field);
      if (is_bit_field)
	{
	  size_node = DECL_SIZE (field);
	  field_type = DECL_RESULT (field);
	}
      else
	{
	  field_type = TREE_TYPE (field);
	  size_node = TYPE_SIZE (field_type);
	  if (DECL_SIZE (field)) {
	    tree alt_size = DECL_SIZE (field);
	    if (TREE_INT_CST_LOW (alt_size) < TREE_INT_CST_LOW (size_node)) {
	      is_bit_field = 1;
	      size_node = DECL_SIZE (field);
	    }
	  }
	}
      bit_offset = TREE_INT_CST_LOW (DECL_FIELD_BITPOS (field));
      size_in_bits = TREE_INT_CST_LOW (size_node);
      fprintf (ffifile, "(\"%s\"", field_name);
      if (is_bit_field)
	fprintf (ffifile, " (bitfield ");
      else
	fprintf (ffifile, " (field ");
      emit_ffi_type_reference (field_type);
      if (is_bit_field)
	fprintf (ffifile, " %d %d", bit_offset, size_in_bits);
      else 
	fprintf (ffifile, " %d %d", bit_offset >> 3, size_in_bits >> 3);
      fprintf (ffifile, "))");
      if (count > 0)
	ffi_unindent (1);
      field = TREE_CHAIN (field);
      count++;
    }
  fprintf (ffifile, ")");
  ffi_unindent (1);
}
  
      

void
ffi_emit_type (info, type)
     struct ffi_typeinfo *info;
     tree type;
{
  char *source_file = info->source_file ? info->source_file : "";
  int line_number = info->source_line;
  tree field;

  switch (info->status)
    {
    case FFI_TYPE_STRUCT:
      fprintf (ffifile, "(struct (\"%s\" %d)", source_file, line_number);
      ffi_indent ();
      fprintf (ffifile, "\"%s\"", info->name);
      ffi_unindent (1);
      ffi_emit_field_list (TYPE_FIELDS (type));
      fprintf (ffifile, ")\n");
      break;

    case FFI_TYPE_UNION:
      fprintf (ffifile, "(union (\"%s\" %d)", source_file, line_number);
      ffi_indent ();
      fprintf (ffifile, "\"%s\"", info->name);
      ffi_unindent (1);
      ffi_emit_field_list (TYPE_FIELDS (type));
      fprintf (ffifile, ")\n");
      break;
     
    case FFI_TYPE_ENUM:
      fprintf (ffifile, "(enum (\"%s\" %d)", source_file, line_number);
      ffi_indent ();
      fprintf (ffifile, "\"%s\"", info->name);
      ffi_unindent (1);
      ffi_emit_enum_values (TYPE_VALUES (type));
      fprintf (ffifile, ")\n");
      ffi_emit_enum_idents (TYPE_VALUES (type));
      break;

    default:
      fprintf (stderr, "Unknown type!\n");
      debug_tree (type);
    }
}

void
ffi_define_type (decl, status)
     tree decl;
     enum ffi_typestatus status;
{
  tree type = TREE_TYPE(decl);
  struct ffi_typeinfo *info = find_ffi_type_info (type);

  info->status = status;

  /*  
      fprintf (stderr, "Defining this type\n");
      debug_tree (type);
  */

  switch (status)
    {
    case FFI_TYPE_PRIMITIVE:
      info->name = ffi_primitive_type_name (decl);
      info->source_file = NULL;
      info->source_line = 0;
      break;

    case FFI_TYPE_TYPEDEF:
      info->name = (char *) xstrdup(IDENTIFIER_POINTER (DECL_NAME (decl)));
      info->source_file = (char *) xstrdup (DECL_SOURCE_FILE (decl));
      info->source_line = DECL_SOURCE_LINE (decl);
      ffi_emit_type_decl (info, decl);
      break;


    }
}

void
ffi_define_builtin_types (syms)
     tree syms;
{
  if (syms) 
    {
      ffi_define_builtin_types (TREE_CHAIN (syms));
      if (TREE_CODE (syms) == TYPE_DECL)
	{
	  tree type = TREE_TYPE (syms);
	  if (TYPE_NAME (type)
	      && TREE_CODE (TYPE_NAME (type)) == TYPE_DECL
	      && TYPE_SIZE (type) != NULL_TREE
	      && ! TREE_ASM_WRITTEN (TYPE_NAME (type)))
	    {
	      /*
		fprintf (stderr, "Defining primitive type:\n");
		debug_tree (type);
		fprintf (stderr, " -- primitive type name = \n");
		debug_tree (TYPE_NAME (type));
	      */
	      ffi_define_type (TYPE_NAME (type), FFI_TYPE_PRIMITIVE);
	    }
	}
    }
}


void
ffi_init (ffi_file, input_file_name, syms)
     FILE *ffi_file;
     char *input_file_name;
     tree syms;
{
  ffi_typevec_len = 100;
  ffi_typevec = (struct ffi_typeinfo *) xmalloc (ffi_typevec_len * sizeof ffi_typevec[0]);
  bzero ((char *) ffi_typevec, ffi_typevec_len * sizeof ffi_typevec[0]);
  if (ffi_file) {
    ffifile = ffi_file;
  } else {
    ffifile = stderr;
  }
  ffi_current_source_file = (char *) xstrdup (input_file_name);
  ffi_define_builtin_types (syms);
  ffi_initialized = 1;
}


void
ffi_define_variable (decl)
     tree decl;
{
  tree type = TREE_TYPE (decl);

  ffi_herald (decl, "var");
  ffi_indent ();
  fprintf (ffifile, "\"%s\"", IDENTIFIER_POINTER (DECL_NAME (decl)));
  ffi_unindent (1);
  ffi_indent ();
  emit_ffi_type_reference (type);
  if (DECL_EXTERNAL (decl))
    fprintf (ffifile, " (extern))\n");
  else
    fprintf (ffifile, " (static))\n");
  ffi_unindent (1);
}

void
ffi_define_function (decl)
     tree decl;
{
  tree type = TREE_TYPE (decl);

  ffi_herald (decl, "function");
  ffi_indent ();
  fprintf (ffifile, "\"%s\"", IDENTIFIER_POINTER (DECL_NAME (decl)));
  ffi_unindent (1);
  ffi_indent ();
  emit_ffi_type_reference (TYPE_MAIN_VARIANT(type));
  if (DECL_EXTERNAL (decl))
    fprintf (ffifile, " (extern))\n");
  else
    fprintf (ffifile, " (static))\n");
  ffi_unindent (1);
}

void
ffi_rest_of_decl_compilation (decl, asmspec, top_level, at_end)
     tree decl;
     const char *asmspec;
     int top_level;
     int at_end;
{
  tree type = TREE_TYPE (decl);

  if (! ffi_initialized) 
    return;

#if 0
  fprintf(stderr, "ffi_rest_of_decl_compilation:\n");
  debug_tree(decl); 
  fprintf(stderr, "\n");
#endif


  switch (TREE_CODE (decl))
    {
    case TYPE_DECL:
      ffi_define_type (decl, FFI_TYPE_TYPEDEF);
      break;

    case FUNCTION_DECL:
      ffi_define_function (decl);
      break;

    case VAR_DECL:
      if (DECL_SOURCE_LINE (decl))
	ffi_define_variable (decl);
      break;

    case PARM_DECL:
    case RESULT_DECL:
      fprintf(stderr, " ??? Parm or Result \n");
      debug_tree (decl);
      exit (1);
      break;

    case CONST_DECL:
      fprintf(stderr, " --- Maybe an enum ?\n");
      debug_tree (decl);
      exit (1);
      break;


    default:
      fprintf(stderr, " ??? Unknown \n");
      debug_tree (decl);
      exit (1);
      break;
    }
}

void
ffi_rest_of_type_compilation (type, top_level)
     tree type;
     int top_level;
{
  if (! ffi_initialized)
    return;
  
  ffi_emit_type (ffi_create_type_info (type), type);
}


/*
  Emit the macro name (possibly including a parenthesized arglist).
  Surround whatever's emitted here with double quotes.
  Return a pointer to the first character of the (possibly empty) definition.
*/
char *
ffi_emit_macro_name (p)
     char *p;
{
  int in_arglist = 0;
  char c;

  putc ('\"', ffifile);
  while (c = *p)
    {
      switch (c)
	{
	case '(':
	  in_arglist = 1;
	  /* Fall down */
	case ',':
	  putc (' ', ffifile);
	  putc (c, ffifile);
	  putc (' ', ffifile);
	  break;

	case ')':
	  putc (' ', ffifile);
	  putc (c, ffifile);
	  putc ('\"', ffifile);
	  
	  while (isspace (*++p));
	  return p;
	
	default:
	  if (isspace (c) && !in_arglist)
	    {
	      putc ('\"', ffifile);
	      while (isspace (*++p));
	      return p;
	    }
	  putc (c, ffifile);
	}
      ++p;
    }
  putc ('\"', ffifile);
  return p;
}

void
ffi_emit_macro_expansion (p)
     char *p;
{
  char c;
  char *q = p + strlen (p);
  
  while ((q > p) && isspace( q[-1] )) q--;
  *q = '\0';

  fprintf(ffifile, " \"");

  while (c = *p++)
    {
      if (c == '\"')
	fputc ('\\', ffifile);
      fputc (c, ffifile);
    }
  fputc ('\"', ffifile);
}
  

void
ffi_debug_define (lineno, buffer)
     unsigned lineno;
     char *buffer;
{
  if (ffifile) {
    fprintf(ffifile, "(macro (\"%s\" %d) ", ffi_current_source_file, lineno);
    ffi_emit_macro_expansion (ffi_emit_macro_name (buffer));
    fprintf(ffifile, ")\n");
  }
}

void
ffi_debug_undef (lineno, buffer)
     unsigned lineno;
     char *buffer;
{
  fprintf(ffifile, "#undef %s\n", buffer);
}
